@@ -1,5 +1,5 @@
 # orm/base.py
-# Copyright (C) 2005-2018 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2021 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -9,72 +9,86 @@
 
 """
 
-from .. import util, inspection, exc as sa_exc
-from ..sql import expression
-from . import exc
 import operator
 
+from . import exc
+from .. import exc as sa_exc
+from .. import inspection
+from .. import util
+from ..sql import expression
+
+
 PASSIVE_NO_RESULT = util.symbol(
-    'PASSIVE_NO_RESULT',
+    "PASSIVE_NO_RESULT",
     """Symbol returned by a loader callable or other attribute/history
     retrieval operation when a value could not be determined, based
     on loader callable flags.
-    """
+    """,
+)
+
+PASSIVE_CLASS_MISMATCH = util.symbol(
+    "PASSIVE_CLASS_MISMATCH",
+    """Symbol indicating that an object is locally present for a given
+    primary key identity but it is not of the requested class.  The
+    return value is therefore None and no SQL should be emitted.""",
 )
 
 ATTR_WAS_SET = util.symbol(
-    'ATTR_WAS_SET',
+    "ATTR_WAS_SET",
     """Symbol returned by a loader callable to indicate the
     retrieved value, or values, were assigned to their attributes
     on the target object.
-    """
+    """,
 )
 
 ATTR_EMPTY = util.symbol(
-    'ATTR_EMPTY',
-    """Symbol used internally to indicate an attribute had no callable."""
+    "ATTR_EMPTY",
+    """Symbol used internally to indicate an attribute had no callable.""",
 )
 
 NO_VALUE = util.symbol(
-    'NO_VALUE',
+    "NO_VALUE",
     """Symbol which may be placed as the 'previous' value of an attribute,
     indicating no value was loaded for an attribute when it was modified,
     and flags indicated we were not to load it.
-    """
+    """,
 )
 
 NEVER_SET = util.symbol(
-    'NEVER_SET',
+    "NEVER_SET",
     """Symbol which may be placed as the 'previous' value of an attribute
     indicating that the attribute had not been assigned to previously.
-    """
+    """,
 )
 
 NO_CHANGE = util.symbol(
     "NO_CHANGE",
     """No callables or SQL should be emitted on attribute access
     and no state should change
-    """, canonical=0
+    """,
+    canonical=0,
 )
 
 CALLABLES_OK = util.symbol(
     "CALLABLES_OK",
     """Loader callables can be fired off if a value
     is not present.
-    """, canonical=1
+    """,
+    canonical=1,
 )
 
 SQL_OK = util.symbol(
     "SQL_OK",
     """Loader callables can emit SQL at least on scalar value attributes.""",
-    canonical=2
+    canonical=2,
 )
 
 RELATED_OBJECT_OK = util.symbol(
     "RELATED_OBJECT_OK",
     """Callables can use SQL to load related objects as well
     as scalar value attributes.
-    """, canonical=4
+    """,
+    canonical=4,
 )
 
 INIT_OK = util.symbol(
@@ -82,104 +96,116 @@ INIT_OK = util.symbol(
     """Attributes should be initialized with a blank
     value (None or an empty collection) upon get, if no other
     value can be obtained.
-    """, canonical=8
+    """,
+    canonical=8,
 )
 
 NON_PERSISTENT_OK = util.symbol(
     "NON_PERSISTENT_OK",
     """Callables can be emitted if the parent is not persistent.""",
-    canonical=16
+    canonical=16,
 )
 
 LOAD_AGAINST_COMMITTED = util.symbol(
     "LOAD_AGAINST_COMMITTED",
     """Callables should use committed values as primary/foreign keys during a
     load.
-    """, canonical=32
+    """,
+    canonical=32,
 )
 
 NO_AUTOFLUSH = util.symbol(
     "NO_AUTOFLUSH",
     """Loader callables should disable autoflush.""",
-    canonical=64
+    canonical=64,
+)
+
+NO_RAISE = util.symbol(
+    "NO_RAISE",
+    """Loader callables should not raise any assertions""",
+    canonical=128,
 )
 
 # pre-packaged sets of flags used as inputs
 PASSIVE_OFF = util.symbol(
     "PASSIVE_OFF",
     "Callables can be emitted in all cases.",
-    canonical=(RELATED_OBJECT_OK | NON_PERSISTENT_OK |
-               INIT_OK | CALLABLES_OK | SQL_OK)
+    canonical=(
+        RELATED_OBJECT_OK | NON_PERSISTENT_OK | INIT_OK | CALLABLES_OK | SQL_OK
+    ),
 )
 PASSIVE_RETURN_NEVER_SET = util.symbol(
     "PASSIVE_RETURN_NEVER_SET",
     """PASSIVE_OFF ^ INIT_OK""",
-    canonical=PASSIVE_OFF ^ INIT_OK
+    canonical=PASSIVE_OFF ^ INIT_OK,
 )
 PASSIVE_NO_INITIALIZE = util.symbol(
     "PASSIVE_NO_INITIALIZE",
     "PASSIVE_RETURN_NEVER_SET ^ CALLABLES_OK",
-    canonical=PASSIVE_RETURN_NEVER_SET ^ CALLABLES_OK
+    canonical=PASSIVE_RETURN_NEVER_SET ^ CALLABLES_OK,
 )
 PASSIVE_NO_FETCH = util.symbol(
-    "PASSIVE_NO_FETCH",
-    "PASSIVE_OFF ^ SQL_OK",
-    canonical=PASSIVE_OFF ^ SQL_OK
+    "PASSIVE_NO_FETCH", "PASSIVE_OFF ^ SQL_OK", canonical=PASSIVE_OFF ^ SQL_OK
 )
 PASSIVE_NO_FETCH_RELATED = util.symbol(
     "PASSIVE_NO_FETCH_RELATED",
     "PASSIVE_OFF ^ RELATED_OBJECT_OK",
-    canonical=PASSIVE_OFF ^ RELATED_OBJECT_OK
+    canonical=PASSIVE_OFF ^ RELATED_OBJECT_OK,
 )
 PASSIVE_ONLY_PERSISTENT = util.symbol(
     "PASSIVE_ONLY_PERSISTENT",
     "PASSIVE_OFF ^ NON_PERSISTENT_OK",
-    canonical=PASSIVE_OFF ^ NON_PERSISTENT_OK
+    canonical=PASSIVE_OFF ^ NON_PERSISTENT_OK,
 )
 
-DEFAULT_MANAGER_ATTR = '_sa_class_manager'
-DEFAULT_STATE_ATTR = '_sa_instance_state'
-_INSTRUMENTOR = ('mapper', 'instrumentor')
+DEFAULT_MANAGER_ATTR = "_sa_class_manager"
+DEFAULT_STATE_ATTR = "_sa_instance_state"
+_INSTRUMENTOR = ("mapper", "instrumentor")
 
-EXT_CONTINUE = util.symbol('EXT_CONTINUE')
-EXT_STOP = util.symbol('EXT_STOP')
+EXT_CONTINUE = util.symbol("EXT_CONTINUE")
+EXT_STOP = util.symbol("EXT_STOP")
+EXT_SKIP = util.symbol("EXT_SKIP")
 
 ONETOMANY = util.symbol(
-    'ONETOMANY',
-    """Indicates the one-to-many direction for a :func:`.relationship`.
+    "ONETOMANY",
+    """Indicates the one-to-many direction for a :func:`_orm.relationship`.
 
     This symbol is typically used by the internals but may be exposed within
     certain API features.
 
-    """)
+    """,
+)
 
 MANYTOONE = util.symbol(
-    'MANYTOONE',
-    """Indicates the many-to-one direction for a :func:`.relationship`.
+    "MANYTOONE",
+    """Indicates the many-to-one direction for a :func:`_orm.relationship`.
 
     This symbol is typically used by the internals but may be exposed within
     certain API features.
 
-    """)
+    """,
+)
 
 MANYTOMANY = util.symbol(
-    'MANYTOMANY',
-    """Indicates the many-to-many direction for a :func:`.relationship`.
+    "MANYTOMANY",
+    """Indicates the many-to-many direction for a :func:`_orm.relationship`.
 
     This symbol is typically used by the internals but may be exposed within
     certain API features.
 
-    """)
+    """,
+)
 
 NOT_EXTENSION = util.symbol(
-    'NOT_EXTENSION',
+    "NOT_EXTENSION",
     """Symbol indicating an :class:`InspectionAttr` that's
     not part of sqlalchemy.ext.
 
     Is assigned to the :attr:`.InspectionAttr.extension_type`
-    attibute.
+    attribute.
 
-    """)
+    """,
+)
 
 _never_set = frozenset([NEVER_SET])
 
@@ -200,6 +226,7 @@ def _generative(*assertions):
             assertion(self, fn.__name__)
         fn(self, *args[1:], **kw)
         return self
+
     return generate
 
 
@@ -208,9 +235,10 @@ def _generative(*assertions):
 def manager_of_class(cls):
     return cls.__dict__.get(DEFAULT_MANAGER_ATTR, None)
 
+
 instance_state = operator.attrgetter(DEFAULT_STATE_ATTR)
 
-instance_dict = operator.attrgetter('__dict__')
+instance_dict = operator.attrgetter("__dict__")
 
 
 def instance_str(instance):
@@ -225,7 +253,7 @@ def state_str(state):
     if state is None:
         return "None"
     else:
-        return '<%s at 0x%x>' % (state.class_.__name__, id(state.obj()))
+        return "<%s at 0x%x>" % (state.class_.__name__, id(state.obj()))
 
 
 def state_class_str(state):
@@ -236,7 +264,7 @@ def state_class_str(state):
     if state is None:
         return "None"
     else:
-        return '<%s>' % (state.class_.__name__, )
+        return "<%s>" % (state.class_.__name__,)
 
 
 def attribute_str(instance, attribute):
@@ -273,7 +301,7 @@ def object_state(instance):
     Raises :class:`sqlalchemy.orm.exc.UnmappedInstanceError`
     if no mapping is configured.
 
-    Equivalent functionality is available via the :func:`.inspect`
+    Equivalent functionality is available via the :func:`_sa.inspect`
     function as::
 
         inspect(instance)
@@ -311,7 +339,7 @@ def _class_to_mapper(class_or_mapper):
 
 
 def _mapper_or_none(entity):
-    """Return the :class:`.Mapper` for the given class or None if the
+    """Return the :class:`_orm.Mapper` for the given class or None if the
     class is not mapped.
     """
 
@@ -324,19 +352,19 @@ def _mapper_or_none(entity):
 
 def _is_mapped_class(entity):
     """Return True if the given object is a mapped class,
-    :class:`.Mapper`, or :class:`.AliasedClass`.
+    :class:`_orm.Mapper`, or :class:`.AliasedClass`.
     """
 
     insp = inspection.inspect(entity, False)
-    return insp is not None and \
-        not insp.is_clause_element and \
-        (
-            insp.is_mapper or insp.is_aliased_class
-        )
+    return (
+        insp is not None
+        and not insp.is_clause_element
+        and (insp.is_mapper or insp.is_aliased_class)
+    )
 
 
 def _attr_as_key(attr):
-    if hasattr(attr, 'key'):
+    if hasattr(attr, "key"):
         return attr.key
     else:
         return expression._column_as_key(attr)
@@ -344,7 +372,7 @@ def _attr_as_key(attr):
 
 def _orm_columns(entity):
     insp = inspection.inspect(entity, False)
-    if hasattr(insp, 'selectable') and hasattr(insp.selectable, 'c'):
+    if hasattr(insp, "selectable") and hasattr(insp.selectable, "c"):
         return [c for c in insp.selectable.c]
     else:
         return [entity]
@@ -352,8 +380,7 @@ def _orm_columns(entity):
 
 def _is_aliased_class(entity):
     insp = inspection.inspect(entity, False)
-    return insp is not None and \
-        getattr(insp, "is_aliased_class", False)
+    return insp is not None and getattr(insp, "is_aliased_class", False)
 
 
 def _entity_descriptor(entity, key):
@@ -377,13 +404,16 @@ def _entity_descriptor(entity, key):
 
     try:
         return getattr(entity, key)
-    except AttributeError:
-        raise sa_exc.InvalidRequestError(
-            "Entity '%s' has no property '%s'" %
-            (description, key)
+    except AttributeError as err:
+        util.raise_(
+            sa_exc.InvalidRequestError(
+                "Entity '%s' has no property '%s'" % (description, key)
+            ),
+            replace_context=err,
         )
 
-_state_mapper = util.dottedgetter('manager.mapper')
+
+_state_mapper = util.dottedgetter("manager.mapper")
 
 
 @inspection._inspects(type)
@@ -402,14 +432,14 @@ def _inspect_mapped_class(class_, configure=False):
 
 
 def class_mapper(class_, configure=True):
-    """Given a class, return the primary :class:`.Mapper` associated
+    """Given a class, return the primary :class:`_orm.Mapper` associated
     with the key.
 
     Raises :exc:`.UnmappedClassError` if no mapping is configured
     on the given class, or :exc:`.ArgumentError` if a non-class
     object is passed.
 
-    Equivalent functionality is available via the :func:`.inspect`
+    Equivalent functionality is available via the :func:`_sa.inspect`
     function as::
 
         inspect(some_mapped_class)
@@ -422,7 +452,8 @@ def class_mapper(class_, configure=True):
     if mapper is None:
         if not isinstance(class_, type):
             raise sa_exc.ArgumentError(
-                "Class object expected, got '%r'." % (class_, ))
+                "Class object expected, got '%r'." % (class_,)
+            )
         raise exc.UnmappedClassError(class_)
     else:
         return mapper
@@ -430,7 +461,7 @@ def class_mapper(class_, configure=True):
 
 class InspectionAttr(object):
     """A base class applied to all ORM objects that can be returned
-    by the :func:`.inspect` function.
+    by the :func:`_sa.inspect` function.
 
     The attributes defined here allow the usage of simple boolean
     checks to test basic facts about the object returned.
@@ -442,10 +473,12 @@ class InspectionAttr(object):
     here intact for forwards-compatibility.
 
     """
+
     __slots__ = ()
 
     is_selectable = False
-    """Return True if this object is an instance of :class:`.Selectable`."""
+    """Return True if this object is an instance of
+    :class:`_expression.Selectable`."""
 
     is_aliased_class = False
     """True if this object is an instance of :class:`.AliasedClass`."""
@@ -454,7 +487,7 @@ class InspectionAttr(object):
     """True if this object is an instance of :class:`.InstanceState`."""
 
     is_mapper = False
-    """True if this object is an instance of :class:`.Mapper`."""
+    """True if this object is an instance of :class:`_orm.Mapper`."""
 
     is_property = False
     """True if this object is an instance of :class:`.MapperProperty`."""
@@ -471,7 +504,7 @@ class InspectionAttr(object):
 
     .. seealso::
 
-        :attr:`.Mapper.all_orm_descriptors`
+        :attr:`_orm.Mapper.all_orm_descriptors`
 
     """
 
@@ -483,14 +516,13 @@ class InspectionAttr(object):
     """
 
     is_clause_element = False
-    """True if this object is an instance of :class:`.ClauseElement`."""
+    """True if this object is an instance of
+    :class:`_expression.ClauseElement`."""
 
     extension_type = NOT_EXTENSION
     """The extension type, if any.
     Defaults to :data:`.interfaces.NOT_EXTENSION`
 
-    .. versionadded:: 0.8.0
-
     .. seealso::
 
         :data:`.HYBRID_METHOD`
@@ -518,12 +550,10 @@ class InspectionAttrInfo(InspectionAttr):
 
         The dictionary is generated when first accessed.  Alternatively,
         it can be specified as a constructor argument to the
-        :func:`.column_property`, :func:`.relationship`, or :func:`.composite`
+        :func:`.column_property`, :func:`_orm.relationship`, or
+        :func:`.composite`
         functions.
 
-        .. versionadded:: 0.8  Added support for .info to all
-           :class:`.MapperProperty` subclasses.
-
         .. versionchanged:: 1.0.0 :attr:`.MapperProperty.info` is also
            available on extension types via the
            :attr:`.InspectionAttrInfo.info` attribute, so that it can apply
@@ -544,4 +574,5 @@ class _MappedAttribute(object):
     attributes.
 
     """
+
     __slots__ = ()
