@@ -1,5 +1,5 @@
 # orm/instrumentation.py
-# Copyright (C) 2005-2019 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2021 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -42,7 +42,7 @@ _memoized_key_collection = util.group_expirable_memoized_property()
 
 
 class ClassManager(dict):
-    """tracks state information at the class level."""
+    """Tracks state information at the class level."""
 
     MANAGER_ATTR = base.DEFAULT_MANAGER_ATTR
     STATE_ATTR = base.DEFAULT_STATE_ATTR
@@ -137,13 +137,28 @@ class ClassManager(dict):
         :class:`.AssociationProxy`.
 
         """
-        if exclude is None:
-            exclude = set()
-        for supercls in self.class_.__mro__:
-            for key in set(supercls.__dict__).difference(exclude):
-                exclude.add(key)
-                val = supercls.__dict__[key]
-                if isinstance(val, interfaces.InspectionAttr):
+
+        found = {}
+
+        # constraints:
+        # 1. yield keys in cls.__dict__ order
+        # 2. if a subclass has the same key as a superclass, include that
+        #    key as part of the ordering of the superclass, because an
+        #    overridden key is usually installed by the mapper which is going
+        #    on a different ordering
+        # 3. don't use getattr() as this fires off descriptors
+
+        for supercls in self.class_.__mro__[0:-1]:
+            inherits = supercls.__mro__[1]
+            for key in supercls.__dict__:
+                found.setdefault(key, supercls)
+                if key in inherits.__dict__:
+                    continue
+                val = found[key].__dict__[key]
+                if (
+                    isinstance(val, interfaces.InspectionAttr)
+                    and val.is_attribute
+                ):
                     yield key, val
 
     def _get_class_attr_mro(self, key, default=None):
@@ -205,7 +220,7 @@ class ClassManager(dict):
         setattr(self.class_, self.MANAGER_ATTR, self)
 
     def dispose(self):
-        """Dissasociate this manager from its class."""
+        """Disassociate this manager from its class."""
 
         delattr(self.class_, self.MANAGER_ATTR)
 
@@ -469,7 +484,7 @@ class InstrumentationFactory(object):
 
 
 # this attribute is replaced by sqlalchemy.ext.instrumentation
-# when importred.
+# when imported.
 _instrumentation_factory = InstrumentationFactory()
 
 # these attributes are replaced by sqlalchemy.ext.instrumentation
